from builtins import str
from datetime import date, timedelta

def _date(ds):
    """Will convert a given string into a date object. The given date can have the following format:

    today
        Will generate to current date
    YYYYMMDD
        Will generate the date of year (YYYY) month (MM) day (DD)

    :v: Value to be checked
    :returns: Date object

    """
    if ds in ["'today'", 'today']:
        return date.today()
    else:
        ds = ds.strip("'")
        y = int(ds[0:4])
        m = int(ds[4:6])
        d = int(ds[6:8])
    return date(y, m, d)


def _timedelta(s):
    """Will convert a given string into a timedelte object. The given
    date can have the following format:

    HH:MM:SS
        Will generate the timedelta of hours (HH) minutes (MM) and
        seconds (SS)

    :v: Value to be checked
    :returns: Timedelta object

    """
    interval  = s.split(":")
    hours = int(interval[0])
    minutes = int(interval[1])
    seconds = int(interval[2])
    td = timedelta(hours=hours, minutes=minutes, seconds=seconds)
    return td


def _bool(v):
    """Will check if the given value is set and not empty. Empty means:

    * For String: No value or empty String
    * For Lists: No value or empty Lists
    * For Numbers: No value (0 is consideret as valid value)

    >>> _bool(None)
    False
    >>> _bool(False)
    False
    >>> _bool(True)
    True
    >>> _bool('')
    False
    >>> _bool('  ')
    True
    >>> _bool([])
    False
    >>> _bool(0)
    True
    >>> _bool('foo')
    True

    :v: Value to be checked
    :returns: True or False

    """
    if v is None:
        return False
    if isinstance(v, bool):
        return v
    if isinstance(v, str):
        if v == "''":
            return False
        else:
            return bool(v)
    elif isinstance(v, list):
        if len(v) == 0:
            return False
        elif v[0] == '':
            # FIXME: check why empty lists become [''] after parsing.
            # (ti) <2014-10-23 10:56>
            return False
        else:
            return True
    else:
        return bool(str(v))


def _len(v):
    """Will return of the given value. For all values except a list the
    value will be converted into a string first and the length of the
    string will be returned. In case of a None value the length will be
    0.

    >>> _len(1.0)
    3
    >>> _len(1234)
    4
    >>> _len([1,2,3,4,5])
    5
    >>> _len('')
    0
    >>> _len(None)
    0

    :v: Value to be checked
    :returns: True or False

    """
    if isinstance(v, list):
        return len(v)
    elif v is None:
        return 0
    return len(unicode(v))

functions = {
    "date": _date,
    "timedelta": _timedelta,
    "bool": _bool,
    "len": _len
}



from __future__ import division
import operator

class Operator(object):
    """Baseclass for all operators"""

    def eval(self):
        raise NotImplementedError()

class Binary(Operator):
    """Baseclass for Binary operators"""

class Unary(Operator):
    """Baseclass for Unary operators"""

class Div(Binary):
    """Devision operator"""

    def eval(self, a, b):
        """Divided a with b. In case a nd b are integer values return a
        integer values. Otherwise a float value will be returned"""
        if isinstance(b, int):
            return int(a / b)
        else:
            return a / b

class In(Binary):
    """In operator"""

    def eval(self, a, b):
        """Returns true if a is in list b"""
        return a in b

operators = {
    "not": operator.not_,
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": Div().eval,
    "<": operator.lt,
    "<=": operator.le,
    ">=": operator.ge,
    ">": operator.gt,
    "==": operator.eq,
    "!=": operator.ne,
    "and": operator.and_,
    "or": operator.or_,
    "in": In().eval
}



from builtins import object
import logging
from pyparsing import ParseResults
from brabbel.parser import Parser
from brabbel.operators import operators
from brabbel.functions import functions


logging.basicConfig()
log = logging.getLogger(__name__)


class OperandMissmatchError(TypeError):
    pass


def _evaluate_term(op, operand):
        if op is None:
            return operand[0]
        elif op == "not":
            return operators[op](operand[0])
        elif op == "in":
            return operators[op](operand[0], operand[1])
        elif type(operand[0]) == type(operand[1]):
            # Only evalutate the term if both operators of the operand are
            # of the same type.
            return operators[op](operand[0], operand[1])
        else:
            # If the type of the operands are not the same, then raise
            # an expection.
            msg = ("Can not use operands '%s' on operator '%s'. "
                   "Type of operands must be equal"
                   % ((operand[0], type(operand[0]),
                      operand[1], type(operand[1])), op))
            raise OperandMissmatchError(msg)


def _resolve_variable(key, values):
    try:
        value = values[key.strip("$")]
    except KeyError:
        log.warning("Variable %s could not found in the values."
                    % key.strip("$"))
        value = None
    return value


class Expression(object):

    """Docstring for Expression. """

    def __init__(self, expression):
        """Initialise a Expression object

        :expression: String representation of an Expression

        """
        self._expression = expression
        self._expression_tree = Parser().parse(self._expression)

    def evaluate(self, values=None):
        """Returns the result auf the evaluation of the expression.

        :values: Dictionary with key value pairs containing values which
        can be used while evaluation
        :returns: Result of the evaluation

        """
        if values is None:
            values = {}
        return self._evaluate(self._expression_tree, values)

    def _evaluate(self, tree, values):
        operand = []
        op = None
        func = None
        try:
            for element in tree:
                if func:
                    param = element[0]
                    if isinstance(param, str) and param.startswith("$"):
                        param = _resolve_variable(param, values)
                    result = func(param)
                    operand.append(result)
                    func = None
                elif isinstance(element, ParseResults):
                    operand.append(self._evaluate(element, values))
                elif element in list(operators.keys()):
                    op = element
                    # Short curcuiting "and" and "or" opertator
                    if op == "and" and not bool(operand[0]):
                        return False
                    if op == "or" and bool(operand[0]):
                        return True
                elif element in list(functions.keys()):
                    func = functions[element]
                else:
                    if isinstance(element, str) and element.startswith("$"):
                        element = _resolve_variable(element, values)
                    operand.append(element)

                # Preevaluate here and use the result as the first operand.
                if len(operand) == 2:
                    operand = [_evaluate_term(op, operand)]
                    op = None
            return _evaluate_term(op, operand)
        except OperandMissmatchError as ex:
            log.error("Can not evaluate expression '%s': %s"
                      % (self._expression, ex.message))
            raise
        except:
            log.exception("Can not evaluate expression '%s'"
                          % self._expression)
            raise



import logging
from builtins import object
from pyparsing import (
    ParserElement,
    Literal, Word,
    Combine, Group, Optional,
    nums, alphanums, alphas,
    delimitedList,
    operatorPrecedence, opAssoc)


"""
number    :: '0'..'9'+
string    :: '0'..'9''a'..'z''_'+
variable  :: '$' string

"""
log = logging.getLogger(__name__)
ParserElement.enablePackrat()

########################################################################
#                               Helpers                                #
########################################################################


def _str(origString, loc, tokens):
    return unicode(tokens[0])


def _number(origString, loc, tokens):
    try:
        return int(tokens[0])
    except:
        return float(tokens[0])


def _make_list(element=""):
    """Returns a list element

    :element: Parsed element as a string representation of a list
    :returns: List element

    """
    listing = []
    for e in element:
        listing.append(e)
    return [listing]


########################################################################
#                                ATOMS                                 #
########################################################################
lpar = Literal("(")
lbr = Literal("[")
rpar = Literal(")")
rbr = Literal("]")
lquote = Literal("'")
rquote = Literal("'")
number = Combine(Optional("-") + Word(nums + '.')).setParseAction(_number)
# TODO: Remove "-" from list of allowed chars. Is only here for
# compatibility. (None) <2014-10-28 14:04>
variable = Combine("$" + Word(alphanums + "_" + "-" + "."))
string = Combine(lquote.suppress() + Optional(Word(alphanums + "_" + " " + "-" + ":")) + rquote.suppress()).setParseAction(_str)
identifier = Word(alphas + "_")
none = Literal("None").setParseAction(lambda t: False)
true = Literal("True").setParseAction(lambda t: True)
false = Literal("False").setParseAction(lambda t: False)
listing = lbr.suppress() + delimitedList(Optional(string | number)).setParseAction(_make_list) + rbr.suppress()
function = identifier.setResultsName("name") + lpar.suppress() + Group(Optional(delimitedList(number | string | variable | listing | true | false | none))) + rpar.suppress()
atom = listing | number | string | variable | true | false | none | function

########################################################################
#                              Operators                               #
########################################################################

opmapping = {
    " ge ": " >= ",
    " gt ": " > ",
    " lt ": " < ",
    " le ": " <= ",
    " eq ": " == ",
    " ne ": " != "
}

o_not = Literal("not")
o_plus = Literal("+")
o_minus = Literal("-")
o_mul = Literal("*")
o_div = Literal("/")
o_gt = Literal(">")
o_ge = Literal(">=")
o_lt = Literal("<")
o_le = Literal("<=")
o_ne = Literal("!=")
o_eq = Literal("==")
o_and = Literal("and")
o_or = Literal("or")
o_in = Literal("in")

bnf = operatorPrecedence(atom,
                         [(o_not, 1, opAssoc.RIGHT),
                          (o_mul, 2, opAssoc.LEFT),
                          (o_div, 2, opAssoc.LEFT),
                          (o_plus, 2, opAssoc.LEFT),
                          (o_minus, 2, opAssoc.LEFT),
                          (o_gt, 2, opAssoc.LEFT),
                          (o_ge, 2, opAssoc.LEFT),
                          (o_lt, 2, opAssoc.LEFT),
                          (o_le, 2, opAssoc.LEFT),
                          (o_ne, 2, opAssoc.LEFT),
                          (o_eq, 2, opAssoc.LEFT),
                          (o_and, 2, opAssoc.LEFT),
                          (o_or, 2, opAssoc.LEFT),
                          (o_in, 2, opAssoc.LEFT),
                          ])


class Parser(object):

    """Parser class for python expression."""

    def __init__(self):
        """@todo: to be defined1. """
        pass

    def parse(self, expr):
        """Returns the BNF-Tree of the given expression

        :expr: String of the expression
        :returns: Returns the parsed BNF form the the expression

        """
        # Replace operators like gt, lt...
        for op in opmapping:
            expr = expr.replace(op, opmapping[op])
        try:
            return bnf.parseString(expr)
        except Exception:
            log.exception("Error on parsing %s" % expr)
